@@ -30,11 +30,6 @@
 #define DISABLE_DEBUGLOG
 
 
-#ifndef ICONV_CONST
-# define ICONV_CONST
-#endif
-
-
 #include "cgui_p.h"
 #include "i18n_l.h"
 
@@ -62,6 +57,12 @@
 #ifdef HAVE_SIGNAL_H
 # include <signal.h>
 #endif
+#ifdef HAVE_ICONV_H
+# include <iconv.h>
+#endif
+#ifndef ICONV_CONST
+# define ICONV_CONST
+#endif
 
 
 
@@ -235,10 +236,34 @@ int GWEN_Gui_CGui__input(GWEN_UNUSED GWEN_GUI *gui,
 #endif
   int chr;
   unsigned int pos;
+  char *pOutbuf;
   int rv;
 #if HAVE_DECL_SIGPROCMASK
   sigset_t snew, sold;
 #endif
+#ifdef HAVE_ICONV
+#define INBUFSIZE 6
+  char inbuf[INBUFSIZE];
+  iconv_t ic;
+  size_t inLeft;
+  size_t outLeft;
+  size_t done;
+  ICONV_CONST char *pInbuf;
+  char *nextchr;
+
+  nextchr=(char *)GWEN_Gui_GetCharSet(gui);
+  if (!nextchr)
+    nextchr="UTF-8";
+  ic=iconv_open("UTF-8", nextchr);
+  if (ic==(iconv_t)-1) {
+    DBG_ERROR(GWEN_LOGDOMAIN, "Cannot convert from \"%s\" to \"UTF-8\", %s",
+              nextchr, strerror(errno));
+    return GWEN_ERROR_GENERIC;
+  }
+
+  pInbuf=inbuf;
+  outLeft=maxLen;
+#endif
 
   /* if possible, disable echo from stdin to stderr during password
    * entry */
@@ -266,14 +291,58 @@ int GWEN_Gui_CGui__input(GWEN_UNUSED GWEN_GUI *gui,
 #endif
 
   pos=0;
-  rv=0;
+  pOutbuf=buffer;
   for (;;) {
+#ifdef HAVE_ICONV
+    nextchr=inbuf;
+    pInbuf=inbuf;
+    inLeft=0;
+    outLeft=maxLen-pos;
+    do {
+      chr=getchar();
+      if (chr==EOF)
+        break;
+      *nextchr++=chr;
+      inLeft++;
+      done=iconv(ic, &pInbuf, &inLeft, &pOutbuf, &outLeft);
+    } while (done==(size_t)-1 && errno==EINVAL &&
+             nextchr-inbuf<INBUFSIZE);
+
+    if (chr==EOF) {
+      DBG_ERROR(GWEN_LOGDOMAIN, "Unexpected EOF while reading from stdin");
+      rv=GWEN_ERROR_GENERIC;
+      break;
+    }
+    else if (done==(size_t)-1) {
+      if (errno==E2BIG || errno==EILSEQ) {
+        GWEN_Gui_StdPrintf(gui, stderr, "\007");
+        continue;
+      }
+      DBG_ERROR(GWEN_LOGDOMAIN, "Unrecoverable error in conversion: %s",
+                strerror(errno));
+      rv=GWEN_ERROR_GENERIC;
+      break;
+    }
+#else /* HAVE_ICONV */
     chr=getchar();
+    if (chr!=EOF)
+      *pOutbuf++=chr;
+    else {
+      DBG_ERROR(GWEN_LOGDOMAIN, "Unexpected EOF while reading from stdin");
+      rv=GWEN_ERROR_GENERIC;
+      break;
+    }
+#endif /* HAVE_ICONV */
+
     if (chr==GWEN_GUI_CGUI_CHAR_DELETE) {
       if (pos) {
-        pos--;
-        fprintf(stderr, "%c %c", 8, 8);
+        /* Look for the start of the previous UTF-8 character */
+        do
+          pos--;
+        while ((buffer[pos]&0xC0)==0x80 && pos);
+        GWEN_Gui_StdPrintf(gui, stderr, "%c %c", 8, 8);
       }
+      pOutbuf=buffer+pos;
     }
     else if (chr==GWEN_GUI_CGUI_CHAR_ENTER) {
       if (minLen && pos<minLen) {
@@ -297,45 +366,49 @@ int GWEN_Gui_CGui__input(GWEN_UNUSED GWEN_GUI *gui,
 	  }
 	}
 	else {
-          /* too few characters */
-	  fprintf(stderr, "\007");
+	  /* too few characters */
+	  GWEN_Gui_StdPrintf(gui, stderr, "\007");
+	  pOutbuf=buffer+pos;
 	}
       }
       else {
-        fprintf(stderr, "\n");
+        GWEN_Gui_StdPrintf(gui, stderr, "\n");
         buffer[pos]=0;
         rv=0;
         break;
       }
     }
-    else {
-      if (pos<maxLen) {
-        if (chr==GWEN_GUI_CGUI_CHAR_ABORT) {
-          DBG_INFO(GWEN_LOGDOMAIN, "User aborted");
-          rv=GWEN_ERROR_USER_ABORTED;
-          break;
-        }
-        else {
-          if ((flags & GWEN_GUI_INPUT_FLAGS_NUMERIC) &&
-              !isdigit(chr)) {
-            /* bad character */
-            fprintf(stderr, "\007");
-          }
-          else {
-            if (flags & GWEN_GUI_INPUT_FLAGS_SHOW)
-              fprintf(stderr, "%c", chr);
-            else
-              fprintf(stderr, "*");
-            buffer[pos++]=chr;
-            buffer[pos]=0;
-          }
-        }
+    else if (chr==GWEN_GUI_CGUI_CHAR_ABORT) {
+      DBG_INFO(GWEN_LOGDOMAIN, "User aborted");
+      rv=GWEN_ERROR_USER_ABORTED;
+      break;
+    }
+    else if (pOutbuf-buffer<maxLen) {
+      if ((flags & GWEN_GUI_INPUT_FLAGS_NUMERIC) &&
+          !isdigit(chr)) {
+        /* bad character */
+        GWEN_Gui_StdPrintf(gui, stderr, "\007");
+        pOutbuf=buffer+pos;
       }
       else {
-        /* buffer full */
-        fprintf(stderr, "\007");
+        if (flags & GWEN_GUI_INPUT_FLAGS_SHOW) {
+          *pOutbuf=0;
+          GWEN_Gui_StdPrintf(gui, stderr, "%s", buffer+pos);
+        }
+        else
+#ifndef HAVE_ICONV
+          /* Do not print stars for continuation bytes */
+          if ((chr&0xC0)!=0x80)
+#endif
+            GWEN_Gui_StdPrintf(gui, stderr, "*");
+        pos=pOutbuf-buffer;
       }
     }
+    else {
+      /* buffer full */
+      GWEN_Gui_StdPrintf(gui, stderr, "\007");
+      pOutbuf=buffer+pos;
+    }
   } /* for */
 
 #ifdef HAVE_TERMIOS_H
@@ -348,6 +421,9 @@ int GWEN_Gui_CGui__input(GWEN_UNUSED GWEN_GUI *gui,
 
 #if HAVE_DECL_SIGPROCMASK
   sigprocmask(SIG_BLOCK, &sold, 0);
+#endif
+#ifdef HAVE_ICONV
+  iconv_close(ic);
 #endif
   return rv;
 }
@@ -362,22 +438,19 @@ int GWEN_Gui_CGui_MessageBox(GWEN_GUI *gui,
 			     const char *b2,
 			     const char *b3,
 			     GWEN_UNUSED uint32_t guiid) {
-  GWEN_GUI_CGUI *cgui;
   GWEN_BUFFER *tbuf;
   int c;
 
   assert(gui);
-  cgui=GWEN_INHERIT_GETDATA(GWEN_GUI, GWEN_GUI_CGUI, gui);
-  assert(cgui);
 
   tbuf=GWEN_Buffer_new(0, 256, 0, 1);
   GWEN_Gui_GetRawText(gui, text, tbuf);
 
   if (GWEN_Gui_GetFlags(gui) & GWEN_GUI_FLAGS_NONINTERACTIVE) {
     if (GWEN_GUI_MSG_FLAGS_SEVERITY_IS_DANGEROUS(flags)) {
-      fprintf(stderr,
-              "Got the following dangerous message:\n%s\n",
-              GWEN_Buffer_GetStart(tbuf));
+      GWEN_Gui_StdPrintf(gui, stderr,
+                         "Got the following dangerous message:\n%s\n",
+                         GWEN_Buffer_GetStart(tbuf));
       GWEN_Buffer_free(tbuf);
       return 0;
     }
@@ -391,44 +464,44 @@ int GWEN_Gui_CGui_MessageBox(GWEN_GUI *gui,
     }
   }
 
-  fprintf(stderr, "===== %s =====\n", title);
-  fprintf(stderr, "%s\n", GWEN_Buffer_GetStart(tbuf));
+  GWEN_Gui_StdPrintf(gui, stderr, "===== %s =====\n", title);
+  GWEN_Gui_StdPrintf(gui, stderr, "%s\n", GWEN_Buffer_GetStart(tbuf));
   GWEN_Buffer_free(tbuf);
   tbuf=0;
 
   if (b1) {
-    fprintf(stderr, "(1) %s", b1);
+    GWEN_Gui_StdPrintf(gui, stderr, "(1) %s", b1);
     if (b2) {
-      fprintf(stderr, "  (2) %s", b2);
+      GWEN_Gui_StdPrintf(gui, stderr, "  (2) %s", b2);
       if (b3) {
-        fprintf(stderr, "  (3) %s", b3);
+        GWEN_Gui_StdPrintf(gui, stderr, "  (3) %s", b3);
       }
     }
-    fprintf(stderr, "\n");
+    GWEN_Gui_StdPrintf(gui, stderr, "\n");
   }
-  fprintf(stderr, "Please enter your choice: ");
+  GWEN_Gui_StdPrintf(gui, stderr, "Please enter your choice: ");
   for(;;) {
     c=GWEN_Gui_CGui__readCharFromStdin(0);
     if (c==EOF) {
-      fprintf(stderr, "Aborted.\n");
+      GWEN_Gui_StdPrintf(gui, stderr, "Aborted.\n");
       return GWEN_ERROR_USER_ABORTED;
     }
     if (!b1 && c==13)
       return 0;
     if (c=='1' && b1) {
-      fprintf(stderr, "1\n");
+      GWEN_Gui_StdPrintf(gui, stderr, "1\n");
       return 1;
     }
     else if (c=='2' && b2) {
-      fprintf(stderr, "2\n");
+      GWEN_Gui_StdPrintf(gui, stderr, "2\n");
       return 2;
     }
     else if (c=='3' && b3) {
-      fprintf(stderr, "3\n");
+      GWEN_Gui_StdPrintf(gui, stderr, "3\n");
       return 3;
     }
     else {
-      fprintf(stderr, "%c", 7);
+      GWEN_Gui_StdPrintf(gui, stderr, "%c", 7);
     }
   } /* for */
 
@@ -451,36 +524,36 @@ int GWEN_Gui_CGui_InputBox(GWEN_GUI *gui,
   tbuf=GWEN_Buffer_new(0, 256, 0, 1);
   GWEN_Gui_GetRawText(gui, text, tbuf);
 
-  fprintf(stderr, "===== %s =====\n", title);
-  fprintf(stderr, "%s\n", GWEN_Buffer_GetStart(tbuf));
+  GWEN_Gui_StdPrintf(gui, stderr, "===== %s =====\n", title);
+  GWEN_Gui_StdPrintf(gui, stderr, "%s\n", GWEN_Buffer_GetStart(tbuf));
   GWEN_Buffer_free(tbuf);
   tbuf=0;
 
   if (flags & GWEN_GUI_INPUT_FLAGS_CONFIRM) {
-    for (;;) {
-      char *lbuffer=0;
+    char *lbuffer=0;
 
-      lbuffer=(char*)malloc(maxLen);
-      if (!lbuffer) {
-        DBG_ERROR(GWEN_LOGDOMAIN, "Not enough memory for %d bytes", maxLen);
-        return GWEN_ERROR_INVALID;
-      }
-      fprintf(stderr, "Input: ");
+    lbuffer=(char*)malloc(maxLen);
+    if (!lbuffer) {
+      DBG_ERROR(GWEN_LOGDOMAIN, "Not enough memory for %d bytes", maxLen);
+      return GWEN_ERROR_INVALID;
+    }
+    for (;;) {
+      GWEN_Gui_StdPrintf(gui, stderr, "Input: ");
       rv=GWEN_Gui_CGui__input(gui, flags, lbuffer, minLen, maxLen, guiid);
       if (rv) {
         free(lbuffer);
         return rv;
       }
 
-      fprintf(stderr, "Again: ");
+      GWEN_Gui_StdPrintf(gui, stderr, "Again: ");
       rv=GWEN_Gui_CGui__input(gui, flags, buffer, minLen, maxLen, guiid);
       if (rv) {
         free(lbuffer);
         return rv;
       }
       if (strcmp(lbuffer, buffer)!=0) {
-        fprintf(stderr,
-                "ERROR: Entries do not match, please try (again or abort)\n");
+        GWEN_Gui_StdPrintf(gui, stderr,
+                           "ERROR: Entries do not match, please try (again or abort)\n");
       }
       else {
         rv=0;
@@ -488,9 +561,10 @@ int GWEN_Gui_CGui_InputBox(GWEN_GUI *gui,
       }
 
     } /* for */
+    free(lbuffer);
   }
   else {
-    fprintf(stderr, "Input: ");
+    GWEN_Gui_StdPrintf(gui, stderr, "Input: ");
     rv=GWEN_Gui_CGui__input(gui, flags, buffer, minLen, maxLen, guiid);
   }
 
@@ -516,8 +590,8 @@ uint32_t GWEN_Gui_CGui_ShowBox(GWEN_GUI *gui,
   tbuf=GWEN_Buffer_new(0, 256, 0, 1);
   GWEN_Gui_GetRawText(gui, text, tbuf);
 
-  fprintf(stderr, "----- %s -----\n", title);
-  fprintf(stderr, "%s\n", GWEN_Buffer_GetStart(tbuf));
+  GWEN_Gui_StdPrintf(gui, stderr, "===== %s =====\n", title);
+  GWEN_Gui_StdPrintf(gui, stderr, "%s\n", GWEN_Buffer_GetStart(tbuf));
   GWEN_Buffer_free(tbuf);
   tbuf=0;
 
