@@ -642,6 +642,11 @@ GWEN_DATE *GWEN_Date_GetThisMonthEnd(const GWEN_DATE *dt) {
   case 11:
     day=30;
     break;
+
+  default:
+    DBG_ERROR(GWEN_LOGDOMAIN, "Invalid month (%d)", GWEN_Date_GetMonth(dt));
+    abort();
+    break;
   }
   return GWEN_Date_fromGregorian(GWEN_Date_GetYear(dt), GWEN_Date_GetMonth(dt), day);
 }
@@ -651,7 +656,7 @@ GWEN_DATE *GWEN_Date_GetThisMonthEnd(const GWEN_DATE *dt) {
 GWEN_DATE *GWEN_Date_GetThisQuarterYearStart(const GWEN_DATE *dt) {
   int m;
   
-  m=(GWEN_Date_GetMonth(dt)-1)>>2;
+  m=GWEN_Date_GetMonth(dt)>>2;
   switch(m) {
   case 0:
     return GWEN_Date_fromGregorian(GWEN_Date_GetYear(dt), 1, 1);
@@ -671,7 +676,7 @@ GWEN_DATE *GWEN_Date_GetThisQuarterYearStart(const GWEN_DATE *dt) {
 GWEN_DATE *GWEN_Date_GetThisQuarterYearEnd(const GWEN_DATE *dt) {
   int m;
   
-  m=(GWEN_Date_GetMonth(dt)-1)>>2;
+  m=GWEN_Date_GetMonth(dt)>>2;
   switch(m) {
   case 0:
     return GWEN_Date_fromGregorian(GWEN_Date_GetYear(dt), 3, 31);
